from __future__ import annotations

import json
from functools import wraps, partial
from importlib import util
from typing import Callable, NamedTuple, TypeVar, Mapping

from absl import app, flags, logging

from absl_extra.notifier import BaseNotifier

T = TypeVar("T", bound=Callable, covariant=True)
FLAGS = flags.FLAGS

flags.DEFINE_string("task", default="main", help="Name of task to run.")

if util.find_spec("pymongo"):
    from pymongo import MongoClient
else:
    logging.warning("pymongo not installed.")

if util.find_spec("ml_collections"):
    from ml_collections import config_flags
else:
    logging.warning("ml_collections not installed")


class MongoConfig(NamedTuple):
    uri: str
    db_name: str
    collection: str | None = None


class _ExceptionHandlerImpl(app.ExceptionHandler):
    def __init__(self, name: str, notifier: BaseNotifier):
        self.name = name
        self.notifier = notifier

    def handle(self, exception: Exception) -> None:
        self.notifier.notify_job_failed(self.name, exception)


_TASKS = {}


def register_task(
    fn: Callable[[str], None], name: str | Callable[[], str] = "main"
) -> None:
    """

    Parameters
    ----------
    fn
    name

    Returns
    -------

    """
    if isinstance(name, Callable):
        name = name()
    if name in _TASKS:
        raise RuntimeError(f"Can't register 2 tasks with same name: {name}")
    _TASKS[name] = fn


def hook_task(
    task: Callable[[...], None],
    app_name: str,
    task_name: str,
    notifier: BaseNotifier,
    config_file: str | None,
) -> Callable[[...], None]:
    @wraps(task)
    def wrapper(*, db=None):
        logging.info("-" * 50)
        logging.info(
            f"Flags: {json.dumps(flags.FLAGS.flag_values_dict(), sort_keys=True, indent=4)}"
        )
        if util.find_spec("ml_collections") and config_file is not None:
            config = config_flags.DEFINE_config_file("config", default=config_file)
        else:
            config = None

        if config is not None:
            config = config.value
            logging.info(
                f"Config: {json.dumps(dict(config), sort_keys=True, indent=4)}"
            )
        logging.info("-" * 50)
        if util.find_spec("tensorflow") is not None:
            import tensorflow as tf

            logging.info(f"TF decices -> {tf.config.list_logical_devices()}")

        notifier.notify_job_started(f"{app_name}.{task_name}")

        kwargs = {}
        if db is not None:
            kwargs["db"] = db
        if config is not None:
            kwargs["config"] = config

        ret_val = task(**kwargs)
        notifier.notify_job_finished(f"{app_name}.{task_name}")
        return ret_val

    return wrapper


def run(
    app_name: str | Callable[[], str] = "app",
    notifier: BaseNotifier | None = None,
    config_file: str | None = None,
    mongo_config: MongoConfig | Mapping[str, ...] | None = None,
):
    """

    Parameters
    ----------
    app_name
    notifier
    config_file
    mongo_config

    Returns
    -------

    """

    def pseudo_main(cmd, **kwargs) -> None:
        TASKS[FLAGS.task](**kwargs)

    if notifier is None:
        notifier = BaseNotifier()

    if util.find_spec("pymongo") and mongo_config is not None:
        if isinstance(mongo_config, Mapping):
            mongo_config = MongoConfig(**mongo_config)
        db = MongoClient(mongo_config.uri).get_database(mongo_config.db_name)
        if mongo_config.collection is not None:
            db = db.get_collection(mongo_config.collection)
    else:
        db = None

    app.install_exception_handler(_ExceptionHandlerImpl(app_name, notifier))
    global _TASKS
    TASKS = {
        k: hook_task(
            task=v,
            task_name=k,
            notifier=notifier,
            config_file=config_file,
            app_name=app_name,
        )
        for k, v in _TASKS.items()
    }
    app.run(partial(pseudo_main, db=db))



from __future__ import annotations
import functools
import logging
from typing import (
    Callable,
    TypeVar,
    Protocol,
    ContextManager,
    runtime_checkable,
    Type,
    Mapping,
)
from contextlib import contextmanager
import tensorflow as tf
import platform


R = TypeVar("R")


def requires_gpu(
    func: Callable[[...], R], linux_only: bool = False
) -> Callable[[...], R]:
    """
    Fail if function is executing on host without access to GPU(s).
    Useful for early detecting container runtime misconfigurations.

    Parameters
    ----------
    func:
        Function, which needs hardware acceleration.
    linux_only:
        If set to true, will ignore check on non-linux hosts.


    Returns
    -------

    func:
        Function with the same signature as original one.

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> R:
        if linux_only and platform.system() != "linux":
            return func(*args, **kwargs)

        if len(tf.config.list_physical_devices("GPU")) == 0:
            raise RuntimeError("No GPU available.")

        return func(*args, **kwargs)

    return wrapper


@runtime_checkable
class StrategyLike(Protocol):
    def scope(self) -> ContextManager:
        ...


class NoOpStrategy:
    def __init__(self, **kwargs):
        pass

    @contextmanager
    def scope(self) -> ContextManager:
        yield


def make_tpu_strategy() -> StrategyLike:
    """
    Used for testing locally scripts, which them must run on Colab TPUs. Allows to keep the same scripts,
    without changing strategy assignment.
    If running on linux, will try to create TPUStrategy. Otherwise, will return NoOpStrategy.

    Parameters
    ----------
    cluster_resolver_kwargs:
        Kwargs passed to TPUClusterResolver.
    connector_kwargs:
        Kwargs passed to experimental_connect_to_cluster.
    strategy_kwargs:
        Kwargs passed to TPUStrategy.

    Returns
    -------

    strategy: TPUStrategy on Linux, NoOpStrategy for other OS hosts.


    Examples
    -------
    >>> strategy = make_tpu_strategy()
    >>> with strategy.scope():
    >>>     model = make_model(...)
    >>>     model.fit(...)
    """
    if platform.system().lower() != "linux":
        logging.warning("Not running on linux, falling back to NoOpStrategy.")
        return NoOpStrategy()

    tpu = tf.distribute.cluster_resolver.TPUClusterResolver()
    tf.config.experimental_connect_to_cluster(tpu)
    tf.tpu.experimental.initialize_tpu_system(tpu)
    strategy = tf.distribute.TPUStrategy(tpu, experimental_spmd_xla_partitioning=True)
    return strategy


def make_gpu_strategy(
    strategy_cls: Type[StrategyLike] | None = None, force: bool = False, **kwargs
) -> StrategyLike:
    """
    Useful for testing locally scripts, which must run on multiple GPUs, without changing scripts structure.

    Parameters
    ----------
    strategy_cls:
        Optional class of the strategy to use. Can be used to choose between e.g., MirroredStrategy and CentralStorage strategies.
    force:

    kwargs:
        Kwargs passed to strategy class __init__ method.


    Returns
    -------

    strategy:
        StrategyLike object.

    Examples
    -------
    >>> strategy = make_gpu_strategy()
    >>> with strategy.scope():
    >>>     model = make_model(...)
    >>>     model.fit(...)
    """
    gpus = tf.config.list_physical_devices("GPU")
    n_gpus = len(gpus)
    if n_gpus == 0:
        logging.warning("No GPUs found, falling back to NoOpStrategy.")
        return NoOpStrategy()
    if n_gpus == 1:
        if force:
            return tf.distribute.OneDeviceStrategy(gpus[0])
        else:
            return NoOpStrategy()

    if strategy_cls is None:
        strategy_cls = tf.distribute.MirroredStrategy

    return strategy_cls(**kwargs)


def supports_mixed_precision() -> bool:
    """Check if mixed precision is supported by available GPUs."""
    tpus = tf.config.list_logical_devices("TPU")
    if len(tpus) != 0:
        logging.info("Mixed precision OK. You should use mixed_bfloat16 for TPU.")
    gpus = tf.config.list_physical_devices("GPU")
    if len(gpus) == 0:
        return False
    gpu_details_list = [tf.config.experimental.get_device_details(g) for g in gpus]
    for details in gpu_details_list:
        cc = details.get("compute_capability")
        if cc >= (7, 0):
            logging.info("Mixed precision OK. You should use mixed_float16 for GPU.")
            return True
    return False
